(load "codegen.scm")
(load "typecheck.scm")

(define (test-f pred actual expected)
  (when (not (pred actual expected))
    (error #f
	   (format "test failed:\nexpected: ~a\nactual:   ~a"
		   expected actual))))

(define (test . xs) (apply test-f (cons equal? xs)))
(define (test-types . xs) (apply test-f (cons types-equal? xs)))

(define (read-file file)
  (call-with-input-file file
    (lambda (p)
      (let loop ((next (read-char p))
                 (result ""))
        (if (eof-object? next)
            result
            (loop (read-char p) (string-append result (string next))))))))

(define (test-prog prog exit-code)
  (display prog)
  (newline)
  (compile-to-binary prog "/tmp/test-prog" host-os)
  (test (system "/tmp/test-prog") exit-code))

(define (test-prog-stdout prog output)
  (display prog)
  (newline)
  (compile-to-binary prog "/tmp/test-prog" host-os)
  (system "/tmp/test-prog > /tmp/test-output.txt")
  (let ((str (read-file "/tmp/test-output.txt")))
    (test str output)))

(test-types (typecheck '(lambda (x) (+ ((lambda (y) (x y 3)) 5) 2)))
	    '(abs (abs int (abs int int)) int))

					; recursive types

(test-types (substitute '((t1 (abs t1 t10))) 't1) '(abs t1 t10))

(test-types (typecheck '(let ([bar (lambda (y) y)]
			      [foo (lambda (x) (foo (bar #t)))])
			  foo))
	    '(abs bool a))

(test-types (typecheck '(let ([bar (lambda (y) y)]
			[foo (lambda (x) (foo (bar #t)))])
		    bar))
	    '(abs a a))

(test-types (typecheck '(let ([foo 3]
			      [bar (+ foo baz)]
			      [baz (- bar 1)])
			  bar))
	    'int)

(test-types (typecheck '(let ([foo 3]
			      [bar (baz foo)]
			      [baz (lambda (x) x)])
			  baz))
	    '(abs a a))

(test-types (typecheck '(let ([foo 3]
			      [bar (baz foo)]
			      [baz (lambda (x) x)])
			  bar))
	    'int)

(test-prog '(+ 1 2) 3)
(test-prog '((lambda (x) ((lambda (y) (+ x y)) 42)) 100) 142)

(test-prog '(* 10 5) 50)

(test-prog '(let ((x (+ 1 32))
		  (y x))
	      ((lambda (z) (+ 2 z)) (* y x)))
	   67) ; exit code modulos at 256
(test-prog '(if ((lambda (x) (= x 2)) 1) 0 (- 32 1)) 31)

(test-prog-stdout '(if (= 3 2) 1 (let () (print "hello world!") 0)) "hello world!")

(test-prog '((lambda (x y) (+ x y)) 1 2) 3)
(test-prog '((lambda (x) (+ ((lambda (y) (+ y 1)) 3) x)) 2) 6)

					; passing closures about
(test-prog '((lambda (z) ((lambda (x) (x 1)) (lambda (y) (+ z y)))) 2) 3)

					; passing builtins about
(test-prog '((lambda (x) ((lambda (a b) (a b 3)) + x)) 3) 6)
(test-prog '(bool->int ((lambda (x) (x #f)) !)) 1)
(test-prog '((lambda (f) (f #t)) bool->int) 1)
(test-prog-stdout '(let () ((lambda (f) (f "foo")) print) 0) "foo")
(test-prog '((lambda (f) (f 3 3)) (lambda (x y) (bool->int (= x y)))) 1)
(test-prog '(bool->int ((lambda (f) (! (f 2 3))) =)) 1)

					; recursion
(test-prog '(let [(inc (lambda (f n x)
			 (if (= n 0)
			     x
			     (f f (- n 1) (+ x 1)))))]
	      (inc inc 3 2))
	   5)
